// Load modules

var Lab = require('lab');
var Hapi = require('..');


// Declare internals

var internals = {};


// Test shortcuts

var expect = Lab.expect;
var before = Lab.before;
var after = Lab.after;
var describe = Lab.experiment;
var it = Lab.test;


describe('Method', function () {

    it('reuses cached method value', function (done) {

        var gen = 0;
        var method = function (id, next) {

            return next(null, { id: id, gen: gen++ });
        };

        var server = new Hapi.Server(0);
        server.method('test', method, { cache: { expiresIn: 1000 } });

        server.start(function () {

            server.methods.test(1, function (err, result) {

                expect(result.gen).to.equal(0);

                server.methods.test(1, function (err, result) {

                    expect(result.gen).to.equal(0);
                    done();
                });
            });
        });
    });

    it('generates new value after cache drop', function (done) {

        var gen = 0;
        var method = function (id, next) {

            return next(null, { id: id, gen: gen++ });
        };

        var server = new Hapi.Server(0);
        server.method('dropTest', method, { cache: { expiresIn: 1000 } });

        server.start(function () {

            server.methods.dropTest(2, function (err, result) {

                expect(result.gen).to.equal(0);
                server.methods.dropTest.cache.drop(2, function (err) {

                    expect(err).to.not.exist;

                    server.methods.dropTest(2, function (err, result) {

                        expect(result.gen).to.equal(1);
                        done();
                    });
                });
            });
        });
    });

    it('errors on invalid drop key', function (done) {

        var gen = 0;
        var method = function (id, next) {

            return next(null, { id: id, gen: gen++ });
        };

        var server = new Hapi.Server(0);
        server.method('dropErrTest', method, { cache: { expiresIn: 1000 } });

        server.start(function () {

            server.methods.dropErrTest.cache.drop(function () {}, function (err) {

                expect(err).to.exist;
                done();
            });
        });
    });

    it('throws an error when name is not a string', function (done) {

        var fn = function () {

            var server = new Hapi.Server();
            server.method(0, function () { });
        };
        expect(fn).to.throw(Error);
        done();
    });

    it('throws an error when fn is not a function', function (done) {

        var fn = function () {

            var server = new Hapi.Server();
            server.method('user', 'function');
        };
        expect(fn).to.throw(Error);
        done();
    });

    it('throws an error when options is not an object', function (done) {

        var fn = function () {

            var server = new Hapi.Server();
            server.method('user', function () { }, 'options');
        };
        expect(fn).to.throw(Error);
        done();
    });

    it('throws an error when options.generateKey is not a function', function (done) {

        var fn = function () {

            var server = new Hapi.Server();
            server.method('user', function () { }, { generateKey: 'function' });
        };
        expect(fn).to.throw(Error);
        done();
    });

    it('throws an error when options.cache is not valid', function (done) {

        var fn = function () {

            var server = new Hapi.Server({ cache: 'memory' });
            server.method('user', function () { }, { cache: { x: 'y' } });
        };
        expect(fn).to.throw(Error);
        done();
    });

    it('returns a valid result when calling a method without using the cache', function (done) {

        var server = new Hapi.Server();
        server.method('user', function (id, next) { return next(null, { id: id }); });
        server.methods.user(4, function (err, result) {

            expect(result.id).to.equal(4);
            done();
        });
    });

    it('returns a valid result when calling a method when using the cache', function (done) {

        var server = new Hapi.Server(0);
        server.start(function () {

            server.method('user', function (id, str, next) { return next(null, { id: id, str: str }); }, { cache: { expiresIn: 1000 } });
            server.methods.user(4, 'something', function (err, result) {

                expect(result.id).to.equal(4);
                expect(result.str).to.equal('something');
                done();
            });
        });
    });

    it('returns an error result when calling a method that returns an error', function (done) {

        var server = new Hapi.Server();
        server.method('user', function (id, next) { return next(new Error()); });
        server.methods.user(4, function (err, result) {

            expect(err).to.exist;
            done();
        });
    });

    it('returns a different result when calling a method without using the cache', function (done) {

        var server = new Hapi.Server();
        var gen = 0;
        server.method('user', function (id, next) { return next(null, { id: id, gen: ++gen }); });
        server.methods.user(4, function (err, result1) {

            expect(result1.id).to.equal(4);
            expect(result1.gen).to.equal(1);
            server.methods.user(4, function (err, result2) {

                expect(result2.id).to.equal(4);
                expect(result2.gen).to.equal(2);
                done();
            });
        });
    });

    describe('with cache', function () {

        it('returns a valid result when calling a method using the cache', function (done) {

            var server = new Hapi.Server(0, { cache: 'memory' });

            var gen = 0;
            server.method('user', function (id, next) { return next(null, { id: id, gen: ++gen }); }, { cache: { expiresIn: 2000 } });

            server.start(function () {

                var id = Math.random();
                server.methods.user(id, function (err, result1) {

                    expect(result1.id).to.equal(id);
                    expect(result1.gen).to.equal(1);
                    server.methods.user(id, function (err, result2) {

                        expect(result2.id).to.equal(id);
                        expect(result2.gen).to.equal(1);
                        done();
                    });
                });
            });
        });

        it('supports empty key method', function (done) {

            var server = new Hapi.Server(0, { cache: 'memory' });

            var gen = 0;
            var terms = 'I agree to give my house';
            server.method('tos', function (next) { return next(null, { gen: gen++, terms: terms }); }, { cache: { expiresIn: 2000 } });

            server.start(function () {

                server.methods.tos(function (err, result1) {

                    expect(result1.terms).to.equal(terms);
                    expect(result1.gen).to.equal(0);
                    server.methods.tos(function (err, result2) {

                        expect(result2.terms).to.equal(terms);
                        expect(result2.gen).to.equal(0);
                        done();
                    });
                });
            });
        });

        it('returns valid results when calling a method (with different keys) using the cache', function (done) {

            var server = new Hapi.Server(0, { cache: 'memory' });
            var gen = 0;
            server.method('user', function (id, next) { return next(null, { id: id, gen: ++gen }); }, { cache: { expiresIn: 2000 } });
            server.start(function () {
                
                var id1 = Math.random();
                server.methods.user(id1, function (err, result1) {

                    expect(result1.id).to.equal(id1);
                    expect(result1.gen).to.equal(1);
                    var id2 = Math.random();
                    server.methods.user(id2, function (err, result2) {

                        expect(result2.id).to.equal(id2);
                        expect(result2.gen).to.equal(2);
                        done();
                    });
                });
            });
        });

        it('returns new object (not cached) when second key generation fails when using the cache', function (done) {

            var server = new Hapi.Server(0, { cache: 'memory' });
            var id1 = Math.random();
            var gen = 0;
            var method = function (id, next) {

                if (typeof id === 'function') {
                    id = id1;
                }

                return next(null, { id: id, gen: ++gen });
            };

            server.method([{ name: 'user', fn: method, options: { cache: { expiresIn: 2000 } } }]);

            server.start(function () {

                server.methods.user(id1, function (err, result1) {

                    expect(result1.id).to.equal(id1);
                    expect(result1.gen).to.equal(1);

                    server.methods.user(function () { }, function (err, result2) {

                        expect(result2.id).to.equal(id1);
                        expect(result2.gen).to.equal(2);
                        done();
                    });
                });
            });
        });
    });
});